#include <stdio.h>

void classify(const float X[], float result[]);

int main() {
    float X[] = {{example}};
    float result[{{num_outputs}}];
    classify(X, result);
    return 0;
}

void classify(const float X[], float result[]) {
    {{code}}
}
